package ru.dbotthepony.kstarbound.math

import com.google.gson.Gson
import com.google.gson.TypeAdapter
import com.google.gson.TypeAdapterFactory
import com.google.gson.reflect.TypeToken
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kommons.gson.consumeNull
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.kommons.vector.Vector2d
import kotlin.math.absoluteValue

private operator fun Vector2d.compareTo(other: Vector2d): Int {
	var cmp = x.compareTo(other.x)
	if (cmp == 0) cmp = y.compareTo(y)
	return cmp
}

data class Line2d(val a: Vector2d, val b: Vector2d) {
	data class Intersection(val intersects: Boolean, val point: KOptional<Vector2d>, val t: KOptional<Double>, val coincides: Boolean, val glances: Boolean) {
		companion object {
			val EMPTY = Intersection(false, KOptional(), KOptional(), false, false)
		}
	}

	fun difference(): Vector2d {
		return b - a
	}

	fun reverse(): Line2d {
		return Line2d(b, a)
	}

	// original source of this intersection algorithm:
	// https://stackoverflow.com/questions/563198/how-do-you-detect-where-two-line-segments-intersect
	// article "Intersection of two lines in three-space" by Ronald Goldman, published in Graphics Gems, page 304
	fun intersect(other: Line2d, infinite: Boolean = false): Intersection {
		val (c, d) = other

		val ab = difference()
		val cd = other.difference()

		val abCross = a.cross(b)
		val cdCross = c.cross(d)

		val denominator = ab.cross(cd)
		val xNumber = abCross * cd.x - cdCross * ab.x
		val yNumber = abCross * cd.y - cdCross * ab.y

		if (denominator.absoluteValue <= NEAR_ZERO) {
			if (xNumber.absoluteValue <= NEAR_ZERO && yNumber.absoluteValue <= NEAR_ZERO) {
				val intersects = infinite || (a >= c && a <= d) || (c >= a && c <= b)
				var point: Vector2d? = null
				var t = 0.0

				if (intersects) {
					if (infinite) {
						point = Vector2d(Double.NEGATIVE_INFINITY, Double.NEGATIVE_INFINITY)
					} else {
						point = if (a < c) c else a
					}
				}

				if (a < c) {
					if (c.x != a.x) {
						t = (c.x - a.x) / ab.x
					} else {
						t = (c.y - a.y) / ab.y
					}
				} else if (a > d) {
					if (d.x != a.x) {
						t = (d.x - a.x) / ab.x
					} else {
						t = (d.y - a.y) / ab.y
					}
				}

				return Intersection(intersects, KOptional.ofNullable(point), KOptional(t), true, intersects)
			} else {
				return Intersection.EMPTY
			}
		} else {
			val ta = (c - a).cross(cd) / denominator
			val tb = (c - a).cross(ab) / denominator

			val intersects = infinite || (ta in 0.0 .. 1.0 && tb in 0.0 .. 1.0)

			return Intersection(
				intersects = intersects,
				t = KOptional(ta),
				point = KOptional((b - a) * ta + a),
				coincides = false,
				glances = !infinite && intersects && (ta <= NEAR_ZERO || ta >= NEAR_ONE || tb <= NEAR_ZERO || tb >= NEAR_ONE)
			)
		}
	}

	fun project(axis: Vector2d): Double {
		val diff = difference()
		return ((axis.x - a.x) * diff.x + (axis.y - a.y) * diff.y) / diff.lengthSquared
	}

	fun distanceTo(other: Vector2d, infinite: Boolean = false): Double {
		var proj = project(other)

		if (!infinite)
			proj = proj.coerceIn(0.0, 1.0)

		return (other - a + difference() * proj).length
	}

	companion object : TypeAdapterFactory {
		const val NEAR_ZERO = Double.MIN_VALUE * 2.0
		const val NEAR_ONE = 1.0 - NEAR_ZERO

		override fun <T : Any> create(gson: Gson, type: TypeToken<T>): TypeAdapter<T>? {
			if (type.rawType == Line2d::class.java) {
				val pair = gson.getAdapter(TypeToken.getParameterized(Vector2d::class.java, Vector2d::class.java)) as TypeAdapter<Pair<Vector2d, Vector2d>>

				return object : TypeAdapter<Line2d>() {
					override fun write(out: JsonWriter, value: Line2d?) {
						if (value == null)
							out.nullValue()
						else
							pair.write(out, value.a to value.b)
					}

					override fun read(`in`: JsonReader): Line2d? {
						if (`in`.consumeNull())
							return null

						val (a, b) = pair.read(`in`)
						return Line2d(a, b)
					}
				} as TypeAdapter<T>
			}

			return null
		}
	}
}
