package ru.dbotthepony.kstarbound.server

import com.google.gson.JsonObject
import io.netty.channel.ChannelHandlerContext
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kommons.io.ByteKey
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.kstarbound.defs.WarpAction
import ru.dbotthepony.kstarbound.defs.WarpAlias
import ru.dbotthepony.kstarbound.defs.WorldID
import ru.dbotthepony.kstarbound.network.Connection
import ru.dbotthepony.kstarbound.network.ConnectionSide
import ru.dbotthepony.kstarbound.network.ConnectionType
import ru.dbotthepony.kstarbound.network.IServerPacket
import ru.dbotthepony.kstarbound.network.packets.ClientContextUpdatePacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.PlayerWarpResultPacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.ServerDisconnectPacket
import ru.dbotthepony.kstarbound.server.world.ServerWorldTracker
import ru.dbotthepony.kstarbound.server.world.WorldStorage
import ru.dbotthepony.kstarbound.server.world.LegacyWorldStorage
import ru.dbotthepony.kstarbound.server.world.ServerWorld
import java.util.HashMap
import java.util.UUID
import java.util.concurrent.CompletableFuture
import java.util.concurrent.ConcurrentLinkedQueue
import kotlin.properties.Delegates

// serverside part of connection
class ServerConnection(val server: StarboundServer, type: ConnectionType) : Connection(ConnectionSide.SERVER, type) {
	var tracker: ServerWorldTracker? = null
	var worldStartAcknowledged = false
	var returnWarp: WarpAction? = null

	val world: ServerWorld?
		get() = tracker?.world

	// packets which interact with world must be
	// executed on world's thread
	fun enqueue(task: ServerWorld.() -> Unit) = tracker?.enqueue(task)

	lateinit var shipWorld: ServerWorld
		private set

	var uuid: UUID? = null

	init {
		connectionID = server.channels.nextConnectionID()

		rpc.add("team.fetchTeamStatus") {
			JsonObject()
		}
	}

	override fun toString(): String {
		val channel = if (hasChannel) channel.remoteAddress().toString() else "<no channel>"
		val world = tracker?.world?.toString() ?: "<not in world>"
		return "ServerConnection[$nickname $uuid ID=$connectionID channel=$channel / $world]"
	}

	private val shipChunks = HashMap<ByteKey, KOptional<ByteArray>>()
	private val modifiedShipChunks = ObjectOpenHashSet<ByteKey>()
	var shipChunkSource by Delegates.notNull<WorldStorage>()
		private set

	override fun setupLegacy() {
		super.setupLegacy()
		shipChunkSource = LegacyWorldStorage.memory(shipChunks)
	}

	override fun setupNative() {
		super.setupNative()
		shipChunkSource = WorldStorage.EMPTY
	}

	fun receiveShipChunks(chunks: Map<ByteKey, KOptional<ByteArray>>) {
		check(shipChunks.isEmpty()) { "Already has ship chunks" }
		shipChunks.putAll(chunks)
	}

	private var remoteVersion = 0L

	override fun flush() {
		if (isConnected) {
			val entries = rpc.write()

			if (entries != null || modifiedShipChunks.isNotEmpty() || server2clientGroup.upstream.hasChangedSince(remoteVersion)) {
				val (data, version) = server2clientGroup.write(remoteVersion, isLegacy)
				remoteVersion = version

				channel.write(ClientContextUpdatePacket(
					entries ?: listOf(),
					KOptional(modifiedShipChunks.associateWith { shipChunks[it]!! }),
					KOptional(data)))

				modifiedShipChunks.clear()
			}
		}

		super.flush()
	}

	override fun onChannelClosed() {
		playerEntity = null

		super.onChannelClosed()
		server.channels.freeConnectionID(connectionID)
		server.channels.connections.remove(this)
		server.freeNickname(nickname)

		announceDisconnect("Connection to remote host is lost.")

		if (::shipWorld.isInitialized) {
			shipWorld.close()
		}

		if (countedTowardsPlayerCount) {
			countedTowardsPlayerCount = false
			server.channels.decrementPlayerCount()
		}
	}

	private var warpingAllowed = false
	private var pendingWarp: Pair<WarpAction, Boolean>? = null
	private var currentWarpStatus: CompletableFuture<*>? = null

	fun enqueueWarp(destination: WarpAction, deploy: Boolean = false) {
		pendingWarp = destination to deploy
	}

	fun tick() {
		if (!isConnected || !channel.isOpen)
			return

		flush()

		if (currentWarpStatus?.isDone == true)
			currentWarpStatus = null

		if (currentWarpStatus == null && warpingAllowed) {
			val pendingWarp = pendingWarp
			this.pendingWarp = null

			if (pendingWarp != null) {
				val (request, deploy) = pendingWarp
				LOGGER.info("Trying to warp $this to $request")

				val resolve = request.resolve(this)

				if (resolve.isLimbo) {
					send(PlayerWarpResultPacket(false, request, true))
				} else if (tracker?.world?.worldID == resolve) {
					LOGGER.info("$this tried to warp into world they are already in.")
					send(PlayerWarpResultPacket(true, request, false))
				} else {
					val world = server.worlds[resolve]

					if (world == null) {
						send(PlayerWarpResultPacket(false, request, false))
					} else {
						currentWarpStatus = world.acceptClient(this).exceptionally {
							send(PlayerWarpResultPacket(false, request, false))

							if (world == shipWorld) {
								disconnect("ShipWorld refused to accept its owner: $it")
							} else {
								enqueueWarp(returnWarp ?: WarpAlias.OwnShip)
							}
						}
					}
				}
			}
		}
	}

	private var announcedDisconnect = true

	private fun announceDisconnect(reason: String) {
		if (!announcedDisconnect && nickname.isNotBlank()) {
			if (reason.isBlank()) {
				server.chat.systemMessage("Player '$nickname' disconnected")
			} else {
				server.chat.systemMessage("Player '$nickname' disconnected ($reason)")
			}

			announcedDisconnect = true
		}
	}

	override fun disconnect(reason: String) {
		announceDisconnect(reason)

		if (channel.isOpen) {
			// send pending updates
			flush()
		}

		tracker?.remove()
		tracker = null

		if (::shipWorld.isInitialized) {
			shipWorld.close()
		}

		if (channel.isOpen) {
			// say goodbye
			channel.write(ServerDisconnectPacket(reason))
			channel.flush()
			channel.close()
		}
	}

	override fun channelRead(ctx: ChannelHandlerContext, msg: Any) {
		if (msg is IServerPacket) {
			try {
				msg.play(this)
			} catch (err: Throwable) {
				LOGGER.error("Failed to handle serverbound packet $msg", err)
				disconnect("Incoming packet caused an exception: $err")
			}
		} else {
			LOGGER.error("Unknown serverbound packet type $msg")
			disconnect("Unknown serverbound packet type $msg")
		}
	}

	private var countedTowardsPlayerCount = false

	override fun inGame() {
		announcedDisconnect = false
		server.chat.systemMessage("Player '$nickname' connected")
		countedTowardsPlayerCount = true
		server.channels.incrementPlayerCount()

		if (isLegacy) {
			ServerWorld.load(server, shipChunkSource, WorldID.ShipWorld(uuid!!)).thenAccept {
				if (!isConnected || !channel.isOpen) {
					LOGGER.warn("$this disconnected before loaded their ShipWorld")
					it.close()
				} else {
					shipWorld = it
					shipWorld.thread.start()
					enqueueWarp(WarpAlias.OwnShip)
					warpingAllowed = true

					if (server.channels.connections.size == 2) {
						enqueueWarp(WarpAction.Player(server.channels.connections.first().uuid!!))
					}
				}
			}.exceptionally {
				LOGGER.error("Error while initializing shipworld for $this", it)
				disconnect("Error while initializing shipworld for player: $it")
				null
			}
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
