package ru.dbotthepony.kstarbound.server.world

import com.google.gson.JsonObject
import it.unimi.dsi.fastutil.ints.IntArraySet
import it.unimi.dsi.fastutil.longs.Long2ObjectFunction
import it.unimi.dsi.fastutil.longs.Long2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ObjectAVLTreeSet
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kommons.util.IStruct2i
import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.WarpAction
import ru.dbotthepony.kstarbound.defs.WorldID
import ru.dbotthepony.kstarbound.defs.tile.TileDamage
import ru.dbotthepony.kstarbound.defs.tile.TileDamageResult
import ru.dbotthepony.kstarbound.defs.world.WorldStructure
import ru.dbotthepony.kstarbound.defs.world.WorldTemplate
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.network.IPacket
import ru.dbotthepony.kstarbound.network.packets.StepUpdatePacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.PlayerWarpResultPacket
import ru.dbotthepony.kstarbound.server.StarboundServer
import ru.dbotthepony.kstarbound.server.ServerConnection
import ru.dbotthepony.kstarbound.util.AssetPathStack
import ru.dbotthepony.kstarbound.util.ExecutionSpinner
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.IChunkListener
import ru.dbotthepony.kstarbound.world.TileHealth
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.WorldGeometry
import ru.dbotthepony.kstarbound.world.api.ImmutableCell
import ru.dbotthepony.kstarbound.world.entities.AbstractEntity
import java.util.concurrent.CompletableFuture
import java.util.concurrent.CopyOnWriteArrayList
import java.util.concurrent.RejectedExecutionException
import java.util.concurrent.atomic.AtomicBoolean
import java.util.concurrent.atomic.AtomicInteger
import java.util.concurrent.locks.LockSupport
import java.util.concurrent.locks.ReentrantLock
import java.util.function.Consumer
import java.util.function.Predicate
import java.util.function.Supplier
import kotlin.concurrent.withLock

class ServerWorld private constructor(
	val server: StarboundServer,
	template: WorldTemplate,
	val storage: WorldStorage,
	val worldID: WorldID,
) : World<ServerWorld, ServerChunk>(template) {
	init {
		if (server.isClosed)
			throw RuntimeException()

		if (worldID != WorldID.Limbo) {
			if (server.worlds.containsKey(worldID))
				throw IllegalStateException("Duplicate world ID: $worldID")

			server.worlds[worldID] = this
		} else {
			server.limboWorlds.add(this)
		}
	}

	val clients = CopyOnWriteArrayList<ServerWorldTracker>()

	private fun doAcceptClient(client: ServerConnection, action: WarpAction?) {
		if (clients.any { it.client == client })
			throw IllegalStateException("$client is already in $this")

		val start = if (action is WarpAction.Player)
			clients.firstOrNull { it.client.uuid == action.uuid }?.client?.playerEntity?.position
		else if (action is WarpAction.World)
			action.target.resolve(this)
		else
			playerSpawnPosition

		if (start == null) {
			client.send(PlayerWarpResultPacket(false, action!!, true))
			throw IllegalStateException("Not a valid spawn target: $action")
		}

		if (action != null)
			client.send(PlayerWarpResultPacket(true, action, false))

		client.tracker?.remove()
		clients.add(ServerWorldTracker(this, client, start))
	}

	fun acceptClient(player: ServerConnection, action: WarpAction? = null): CompletableFuture<Unit> {
		check(!isClosed.get()) { "$this is invalid" }
		unpause()

		try {
			return CompletableFuture.supplyAsync(Supplier { doAcceptClient(player, action) }, mailbox).exceptionally {
				LOGGER.error("Error while accepting new player into world", it)
			}
		} catch (err: RejectedExecutionException) {
			return CompletableFuture.failedFuture(err)
		}
	}

	val spinner = ExecutionSpinner(mailbox::executeQueuedTasks, ::spin, Starbound.TIMESTEP_NANOS)
	private val str = "Server World ${if (worldID == WorldID.Limbo) "limbo(${server.limboWorldIndex.getAndIncrement()})" else worldID.toString()}"
	val thread = Thread(spinner, str)

	private val isClosed = AtomicBoolean()

	init {
		thread.isDaemon = true
	}

	fun pause() {
		if (!isClosed.get()) spinner.pause()
	}

	fun unpause() {
		if (!isClosed.get()) spinner.unpause()
	}

	override fun toString(): String {
		if (isClosed.get())
			return "NULL $str"
		else
			return str
	}

	override fun close() {
		if (!isClosed.get())
			LOGGER.info("Shutting down $this")

		if (isClosed.compareAndSet(false, true)) {
			super.close()
			spinner.unpause()
			clients.forEach { it.remove() }

			if (worldID != WorldID.Limbo)
				server.worlds.remove(worldID)
			else
				server.limboWorlds.remove(this)

			LockSupport.unpark(thread)
		}
	}

	private fun spin(): Boolean {
		if (isClosed.get()) return false

		try {
			tick()
			return true
		} catch (err: Throwable) {
			LOGGER.fatal("Exception in world tick loop", err)
			close()
			return false
		}
	}

	override val isRemote: Boolean
		get() = false

	override fun isSameThread(): Boolean {
		return Thread.currentThread() === thread
	}

	fun damageTiles(positions: Collection<IStruct2i>, isBackground: Boolean, sourcePosition: Vector2d, damage: TileDamage, source: AbstractEntity? = null): TileDamageResult {
		if (damage.amount <= 0.0)
			return TileDamageResult.NONE

		val actualPositions = positions.stream().map { geometry.wrap(it) }.distinct().toList()
		var topMost = TileDamageResult.NONE

		for (pos in actualPositions) {
			val chunk = chunkMap[geometry.chunkFromCell(pos)] ?: continue
			topMost = topMost.coerceAtLeast(chunk.damageTile(pos - chunk.pos.tile, isBackground, sourcePosition, damage, source))
		}

		return topMost
	}

	override fun tick() {
		super.tick()
		val packet = StepUpdatePacket(ticks)

		clients.forEach {
			if (!isClosed.get()) {
				it.send(packet)

				try {
					it.tick()
				} catch (err: Throwable) {
					LOGGER.error("Exception while ticking player ${it.client}", err)
					//it.disconnect("Exception while ticking player: $err")
				}
			}
		}

		ticketListLock.withLock {
			ticketLists.removeIf {
				val valid = it.tick()

				if (!valid) {
					val removed = ticketMap.remove(it.pos.toLong())
					check(removed == it) { "Expected to remove $it, but removed $removed" }

					val chunk = chunkMap[it.pos]

					if (chunk != null) {
						val unloadable = entityIndex
							.query(
								chunk.aabb,
								filter = Predicate { it.isApplicableForUnloading && chunk.aabb.isInside(it.position) },
								distinct = true, withEdges = false)

						storage.saveCells(it.pos, chunk.copyCells())
						storage.saveEntities(it.pos, unloadable)

						unloadable.forEach {
							it.remove()
						}

						chunkMap.remove(it.pos)
					}
				}

				!valid
			}
		}
	}

	override fun broadcast(packet: IPacket) {
		clients.forEach {
			it.send(packet)
		}
	}

	override fun chunkFactory(pos: ChunkPos): ServerChunk {
		return ServerChunk(this, pos)
	}

	private val ticketMap = Long2ObjectOpenHashMap<TicketList>()
	private val ticketLists = ArrayList<TicketList>()
	private val ticketListLock = ReentrantLock()

	private fun getTicketList(pos: ChunkPos): TicketList {
		return ticketMap.computeIfAbsent(geometry.wrapToLong(pos), Long2ObjectFunction { TicketList(it) })
	}

	fun permanentChunkTicket(pos: ChunkPos): ITicket {
		ticketListLock.withLock {
			return getTicketList(pos).Ticket()
		}
	}

	fun temporaryChunkTicket(pos: ChunkPos, time: Int): ITicket {
		require(time > 0) { "Invalid ticket time: $time" }

		ticketListLock.withLock {
			return getTicketList(pos).TimedTicket(time)
		}
	}

	override fun onChunkCreated(chunk: ServerChunk) {
		ticketMap[chunk.pos.toLong()]?.let { chunk.addListener(it) }
	}

	override fun onChunkRemoved(chunk: ServerChunk) {
		ticketMap[chunk.pos.toLong()]?.let { chunk.removeListener(it) }
	}

	interface ITicket {
		fun cancel()
		val isCanceled: Boolean
		val pos: ChunkPos
		val id: Int
		val chunk: ServerChunk?
		var listener: IChunkListener?
	}

	interface ITimedTicket : ITicket, Comparable<ITimedTicket> {
		val timeRemaining: Int
		fun prolong(ticks: Int)

		override fun compareTo(other: ITimedTicket): Int {
			val cmp = timeRemaining.compareTo(other.timeRemaining)
			if (cmp != 0) return cmp
			return id.compareTo(other.id)
		}
	}

	private inner class TicketList(val pos: ChunkPos) : IChunkListener {
		constructor(pos: Long) : this(ChunkPos(pos))

		private var first = true
		private val permanent = ArrayList<Ticket>()
		private val temporary = ObjectAVLTreeSet<TimedTicket>()
		private var ticks = 0
		private var nextTicketID = AtomicInteger()

		val isValid: Boolean
			get() = temporary.isNotEmpty() || permanent.isNotEmpty()

		fun tick(): Boolean {
			ticks++

			while (temporary.isNotEmpty() && temporary.first().timeRemaining <= 0) {
				val ticket = temporary.first()
				ticket.isCanceled = true
				temporary.remove(ticket)
			}

			return temporary.isNotEmpty() || permanent.isNotEmpty()
		}

		override fun onCellChanges(x: Int, y: Int, cell: ImmutableCell) {
			permanent.forEach { it.listener?.onCellChanges(x, y, cell) }
			temporary.forEach { it.listener?.onCellChanges(x, y, cell) }
		}

		override fun onTileHealthUpdate(x: Int, y: Int, isBackground: Boolean, health: TileHealth) {
			permanent.forEach { it.listener?.onTileHealthUpdate(x, y, isBackground, health) }
			temporary.forEach { it.listener?.onTileHealthUpdate(x, y, isBackground, health) }
		}

		abstract inner class AbstractTicket : ITicket {
			final override val id: Int = nextTicketID.getAndIncrement()
			final override val pos: ChunkPos
				get() = this@TicketList.pos

			final override var isCanceled: Boolean = false
			private var loadFuture: CompletableFuture<*>? = null

			fun init() {
				if (first) {
					first = false

					if (geometry.x.inBoundsChunk(pos.x) && geometry.y.inBoundsChunk(pos.y)) {
						ticketListLock.withLock {
							ticketLists.add(this@TicketList)
						}

						val existing = chunkMap[pos]

						if (existing == null) {
							loadFuture = storage.loadCells(pos).thenAccept { tiles ->
								if (!tiles.isPresent) return@thenAccept

								storage.loadEntities(pos).thenAcceptAsync(Consumer { ents ->
									val chunk = chunkMap.compute(pos) ?: return@Consumer
									chunk.loadCells(tiles.value)

									ents.ifPresent {
										for (obj in it) {
											obj.joinWorld(this@ServerWorld)
										}
									}
								}, mailbox)
							}
						} else {
							existing.addListener(this@TicketList)
						}
					}
				} else {
					chunkMap[pos]?.addListener(this@TicketList)
				}
			}

			final override fun cancel() {
				if (isCanceled) return

				ticketListLock.withLock {
					if (isCanceled) return
					isCanceled = true
					loadFuture?.cancel(false)
					listener = null
					cancel0()
				}
			}

			protected abstract fun cancel0()
			final override val chunk: ServerChunk?
				get() = chunkMap[pos]

			final override var listener: IChunkListener? = null
		}

		inner class Ticket : AbstractTicket() {
			init {
				permanent.add(this)
				init()
			}

			override fun cancel0() {
				permanent.remove(this)
			}
		}

		inner class TimedTicket(expiresAt: Int) : AbstractTicket(), ITimedTicket {
			var expiresAt = expiresAt + ticks

			override val timeRemaining: Int
				get() = (expiresAt - ticks).coerceAtLeast(0)

			init {
				temporary.add(this)
				init()
			}

			override fun cancel0() {
				temporary.remove(this)
			}

			override fun prolong(ticks: Int) {
				if (ticks == 0 || isCanceled) return

				ticketListLock.withLock {
					if (isCanceled) return

					temporary.remove(this)
					expiresAt += ticks
					if (timeRemaining > 0) temporary.add(this)
				}
			}
		}
	}

	@JsonFactory
	data class MetadataJson(
		val playerStart: Vector2d,
		val respawnInWorld: Boolean,
		val adjustPlayerStart: Boolean,
		val worldTemplate: JsonObject,
		val centralStructure: WorldStructure,
		val protectedDungeonIds: IntArraySet,
		val worldProperties: JsonObject,
		val spawningEnabled: Boolean
	)

	companion object {
		private val LOGGER = LogManager.getLogger()

		fun create(server: StarboundServer, template: WorldTemplate, storage: WorldStorage, worldID: WorldID = WorldID.Limbo): ServerWorld {
			return ServerWorld(server, template, storage, worldID)
		}

		fun create(server: StarboundServer, geometry: WorldGeometry, storage: WorldStorage, worldID: WorldID = WorldID.Limbo): ServerWorld {
			return ServerWorld(server, WorldTemplate(geometry), storage, worldID)
		}

		fun load(server: StarboundServer, storage: WorldStorage, worldID: WorldID = WorldID.Limbo): CompletableFuture<ServerWorld> {
			LOGGER.info("Attempting to load world at $worldID")

			return storage.loadMetadata().thenApply {
				LOGGER.info("Loading world at $worldID")
				AssetPathStack("/") { _ ->
					val meta = it.map { Starbound.gson.fromJson(it.data.content, MetadataJson::class.java) }.orThrow { NoSuchElementException("No world metadata is present") }

					val world = ServerWorld(server, WorldTemplate.fromJson(meta.worldTemplate), storage, worldID)
					world.playerSpawnPosition = meta.playerStart
					world.respawnInWorld = meta.respawnInWorld
					world.adjustPlayerSpawn = meta.adjustPlayerStart
					world.centralStructure = meta.centralStructure
					world.protectedDungeonIDs.addAll(meta.protectedDungeonIds)
					world
				}
			}.exceptionally {
				LOGGER.error("Error while instancing world $worldID", it)
				null
			}
		}
	}
}
