package ru.dbotthepony.kstarbound.world

import ru.dbotthepony.kommons.io.readVector2d
import ru.dbotthepony.kommons.io.readVector2f
import ru.dbotthepony.kommons.io.writeStruct2d
import ru.dbotthepony.kommons.io.writeStruct2f
import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.tile.TileDamage
import ru.dbotthepony.kstarbound.defs.tile.TileDamageConfig
import ru.dbotthepony.kstarbound.defs.tile.TileDamageType
import java.io.DataInputStream
import java.io.DataOutputStream

class TileHealth() {
	constructor(stream: DataInputStream, isLegacy: Boolean) : this() {
		read(stream, isLegacy)
	}

	var isHarvested: Boolean = false
		private set
	var damageSource: Vector2d = Vector2d.ZERO
		private set
	var damageType: TileDamageType = TileDamageType.PROTECTED
		private set
	var damagePercent: Double = 0.0
		private set
	var damageEffectTimeFactor: Double = 0.0
		private set
	var damageEffectPercentage: Double = 0.0
		private set

	fun copy(): TileHealth {
		val copy = TileHealth()

		copy.isHarvested = isHarvested
		copy.damageSource = damageSource
		copy.damageType = damageType
		copy.damagePercent = damagePercent
		copy.damageEffectTimeFactor = damageEffectTimeFactor
		copy.damageEffectPercentage = damageEffectPercentage

		return copy
	}

	val isHealthy: Boolean
		get() = damagePercent <= 0.0

	val isDead: Boolean
		get() = damagePercent >= 1.0 && damageType != TileDamageType.PROTECTED

	fun write(stream: DataOutputStream, isLegacy: Boolean) {
		if (isLegacy) {
			stream.writeFloat(damagePercent.toFloat())
			stream.writeFloat(damageEffectTimeFactor.toFloat())
			stream.writeBoolean(isHarvested)
			stream.writeStruct2f(damageSource.toFloatVector())
		} else {
			stream.writeDouble(damagePercent)
			stream.writeDouble(damageEffectTimeFactor)
			stream.writeBoolean(isHarvested)
			stream.writeStruct2d(damageSource)
		}

		stream.writeByte(damageType.ordinal)
	}

	fun read(stream: DataInputStream, isLegacy: Boolean) {
		if (isLegacy) {
			damagePercent = stream.readFloat().toDouble()
			damageEffectTimeFactor = stream.readFloat().toDouble()
			isHarvested = stream.readBoolean()
			damageSource = stream.readVector2f().toDoubleVector()
		} else {
			damagePercent = stream.readDouble()
			damageEffectTimeFactor = stream.readDouble()
			isHarvested = stream.readBoolean()
			damageSource = stream.readVector2d()
		}

		damageType = TileDamageType.entries[stream.readUnsignedByte()]
		damageEffectPercentage = damageEffectTimeFactor.coerceIn(0.0, 1.0) * damagePercent
	}

	fun reset() {
		isHarvested = false
		damageSource = Vector2d.ZERO
		damageType = TileDamageType.PROTECTED
		damagePercent = 0.0
		damageEffectTimeFactor = 0.0
	}

	fun damage(config: TileDamageConfig, source: Vector2d, damage: TileDamage) {
		val actualDamage = config.damageDone(damage) / config.totalHealth
		damagePercent = (damagePercent + actualDamage).coerceAtMost(1.0)
		isHarvested = damage.harvestLevel >= config.harvestLevel
		damageSource = source
		damageType = damage.type

		if (actualDamage > 0.0)
			damageEffectTimeFactor = config.maximumEffectTime

		damageEffectPercentage = damageEffectTimeFactor.coerceIn(0.0, 1.0) * damagePercent
	}

	fun tick(config: TileDamageConfig, delta: Double = Starbound.TIMESTEP): Boolean {
		if (isDead || isHealthy)
			return false

		damagePercent -= config.damageRecovery * delta / config.totalHealth
		damageEffectTimeFactor -= delta

		if (damagePercent <= 0.0) {
			damagePercent = 0.0
			damageEffectTimeFactor = 0.0
			damageType = TileDamageType.PROTECTED
		}

		damageEffectPercentage = damageEffectTimeFactor.coerceIn(0.0, 1.0) * damagePercent
		return damagePercent > 0.0
	}
}
